// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#include "PhysicsObject.h"

#ifndef FFloat_h_included
#include "FFloat.h"
#endif

#ifndef Vec3_h_included
#include "Vec3.h"
#endif

#ifndef BoundingSphere_h_included
#include "BoundingSphere.h"
#endif

#ifndef BoundingBox_h_included
#include "BoundingBox.h"
#endif

#ifndef Configuration_h_included
#include "Configuration.h"
#endif

#ifndef string_h_included
#include "string.h"
#define string_h_included
#endif

#ifndef SceneGraphPOInterface_h_included
#include "SceneGraphPOInterface.h"
#endif

PhysicsObject::PhysicsObject (const FString &objectDrawerClass,
			      SP<Configuration> conf)
  : ObjectDrawerConnection (objectDrawerClass, conf)
{
}

static inline FFloat square (FFloat val) {
  return val * val;
}

BoundingSphere PhysicsObject::physicsBoundingSphere (const FFloat *state) const
{
  return BoundingSphere (physicsBoundingSphereRadius(),
			 boundingSphereCenter (state));
}

BoundingBox PhysicsObject::physicsBoundingBox (const FFloat *state) const {
  return BoundingBox (physicsBoundingSphere (state));
}

void PhysicsObject::doPhysics (const FFloat *state, 
			       FFloat *deriv,
			       int myIndex,
			       const Dynavec<PhysicsObjectInfo> &neighbors)
  const
{
  // Suppress warnings.
  (void) state; (void) deriv; (void) myIndex; (void) neighbors;
  // Does nothing.
  return;
}

void PhysicsObject::doPhysics (const FFloat *state, 
			       FFloat *deriv,
			       int myIndex,
			       const Dynavec<PhysicsObjectInfo> &neighbors,
			       const Dynavec<PhysicsObjectInfo> &links)
  const
{
  (void) links;
  doPhysics (state, deriv, myIndex, neighbors);
  return;
}

int PhysicsObject::doPhysics (const SceneGraphPOInterface *sg,
                              const FFloat *state,
                              FFloat *deriv,
                              int myObjectIndex) const {
  int startIndex = sg->objectStateIndex (myObjectIndex);
  doPhysics (state, deriv, startIndex,
             sg->neighbors (myObjectIndex),
             sg->links (myObjectIndex));
  return myObjectIndex + 1;
}

void PhysicsObject::translate (Dynavec <FFloat> &state, const Vec3 &v) const {
  assert (state.size() == stateSize());
  translate (&(state[0]), v);
}

void PhysicsObject::addVelocity (Dynavec <FFloat> &state, const Vec3 &v) const {
  assert (state.size() == stateSize());
  addVelocity (&(state[0]), v);
}

bool PhysicsObject::hasAnimationEver () const {
  return false;
}

bool PhysicsObject::hasAnimationNow (const SceneGraph *sg, int index) const {
  (void) sg; (void) index;
  return false;
}

void PhysicsObject::animateTimeStep (SceneGraph *sg, int index) const {
  (void) sg; (void) index;
}

void PhysicsObject::animateFrame (Canvas *canvas, SceneGraph *sg, int index)
  const
{
  (void) canvas; (void) sg; (void) index;
}

bool PhysicsObject::isSelectable () const {
  return true;
}

bool PhysicsObject::isRelevantNeighbor (const PhysicsObject *o, int myIndex,
					int neighborIndex) const
{
  (void) o;
  (void) myIndex;
  (void) neighborIndex;
  return true;
}

bool PhysicsObject::lessThan (const PhysicsObject *other) const {
  const char *thisKey = keyForCompare ();
  const char *thatKey = other->keyForCompare ();
  return compareKeyLessThan (thisKey, thatKey);
}

bool PhysicsObject::compareKeyLessThan
(const char *thisKey, const char *otherKey)
{
  const int i = strcmp (thisKey, otherKey);
  if (i < 0) {
    return true;
  } else if (i > 0) {
    return false;
  } else {
    return thisKey < otherKey;
  }
}

const char *PhysicsObject::keyForCompare () const {
  return classId ();
}
