// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#include "TopConfiguration.h"

#ifndef VectorSceneConfiguration_h_included
#include "VectorSceneConfiguration.h"
#endif

#ifndef VectorSlotValue_h_included
#include "VectorSlotValue.h"
#endif

#ifndef TestSceneConfiguration_h_included
#include "TestSceneConfiguration.h"
#endif

#ifndef SceneLoader_h_included
#include "SceneLoader.h"
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

#ifndef Configuration_h_included
#include "Configuration.h"
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef DiskConfiguration_h_included
#include "DiskConfiguration.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef RecursiveSlotValue_h_included
#include "RecursiveSlotValue.h"
#endif

#ifndef PositiveIntSlotValue_h_included
#include "PositiveIntSlotValue.h"
#endif

#ifndef ConstantRecursiveSlotValue_h_included
#include "ConstantRecursiveSlotValue.h"
#endif

#ifndef Vec3_h_included
#include "Vec3.h"
#endif

#ifndef ConstantRecursiveSlotValue_h_included
#include "ConstantRecursiveSlotValue.h"
#endif

#ifndef PositiveFloatSlotValue_h_included
#include "PositiveFloatSlotValue.h"
#endif

namespace {
  // Don't have a static variable with a string, it looks like a memory leak.
  // Instead use no-argument functions.
  const FString sceneGraphName () {
    return "Scene Graph";
  }

  const FString sceneName () {
    return "Scene Generator";
  }

  const FString stepsPerFrameName () {
    return "Timesteps Per Frame";
  }

  const FString timePerStepName () {
    return "Time Per Step";
  }

  const FString toplevelName () {
    return "Top Level";
  }

  const FString objectDrawerName () {
    return "Object Drawer";
  }

  FString integratorName () {
    return "Integration Method";
  }

  SP<RecursiveSlotValue> makeTestConf (const FString &className,
				       SP<Configuration> objConf,
				       const Vec3 &translation,
				       int repeat,
				       FFloat spacing)
  {
    SP<RecursiveSlotValue> result = NEW (RecursiveSlotValue ("SceneLoader",
							     "TestSceneLoader"));
    SP<TestSceneConfiguration> sceneConf =
      dynamic_cast <TestSceneConfiguration *> (result->getConfiguration());
    assert (sceneConf);
    SP<RecursiveSlotValue> rsv = NEW (RecursiveSlotValue ("PhysicsObject",
							  className));
    rsv->setConfiguration (objConf);
    sceneConf->setRecursiveSlotValue (rsv);
    sceneConf->setWhere (translation);
    sceneConf->setCubeSize (repeat);
    sceneConf->setCubeSpace (spacing);
    return &*result;
  }
  
  static SP<RecursiveSlotValue> diskAt (const Vec3 &where,
					const Vec3 &normal)
  {
    SP<DiskConfiguration> di = NEW (DiskConfiguration ());
    di->setDefaultNorm (normal);
    return makeTestConf ("Disk", &*di, where, 1, 1);
  }

  static SP<RecursiveSlotValue> cubeAt (const FString &name,
					const Vec3 &where, int size,
					FFloat spacing)
  {
    SP<Factory> f = FactoryTable::load ("PhysicsObject", name);
    assert (f);
    return makeTestConf (name, f->defaultConfiguration(),
			 where, size, spacing);
  }

  static SP<RecursiveSlotValue> initialSceneLoader () {
    SP<RecursiveSlotValue> result =
      NEW (RecursiveSlotValue ("SceneLoader", "VectorSceneLoader"));
    SP<Configuration> conf = result->getConfiguration ();
    assert (conf);
    SP<VectorSceneConfiguration> vsc =
      dynamic_cast <VectorSceneConfiguration *> (&*conf);
    assert (vsc);
    SP<VectorSlotValue> vsv = vsc->getVectorSlotValue ();
    vsv->extendTo (6);
    vsv->setValue (0, &*diskAt (Vec3 ( 0, 5,0), Vec3 (0,-1,1)));
    vsv->setValue (1, &*diskAt (Vec3 ( 0,-5,0), Vec3 (0, 1,1)));
    vsv->setValue (2, &*diskAt (Vec3 ( 5, 0,0), Vec3 (-1,0,1)));
    vsv->setValue (3, &*diskAt (Vec3 (-5, 0,0), Vec3 ( 1,0,1)));
    vsv->setValue (4, &*cubeAt ("RedBouncingBall",Vec3(0,0,5),
				8 ,1));
    vsv->setValue (5, &*cubeAt ("BlueBouncingBall", Vec3(0,0,15), 3, 2));
    return result;
  }
} // namespace

TopConfiguration::TopConfiguration (const FString &topLevelClass)
{
  initializeSlot
    (sceneGraphName (),
     NEW (RecursiveSlotValue ("SceneGraph", "SubsetStorageScene")));
  initializeSlot (sceneName (), &*(initialSceneLoader ()));
  initializeSlot (stepsPerFrameName (), NEW (PositiveIntSlotValue (5)));
  initializeSlot (timePerStepName (), NEW (PositiveFloatSlotValue (0.5)));
  initializeSlot (toplevelName (),
		  NEW (RecursiveSlotValue ("TopLevel", topLevelClass)));
  // AB4 leaks hydrogen through the barrel of the bucky gun!
  // AB2 does too!
  // AB3 at a timestep of 0.5 fs does the buckygun in 3 minutes, 30 seconds.
  // RungeKutta at a timestep of 0.5 fs does the buckygun in 7 minutes.
  // I expected AB3 to be four times faster than RungeKutta, and it's only two;
  // what's going on here?  Maybe we spend a lot of time neighboring.
  initializeSlot (integratorName (),
                  NEW (RecursiveSlotValue ("Integrator", "AB3")));
}

RecursiveSlotValue *
TopConfiguration::getSceneGraphSlot () {
  return RecursiveSlotValue::getSlotValue (this, sceneGraphName ());
}

RecursiveSlotValue *
TopConfiguration::getSceneLoaderSlot () {
  return RecursiveSlotValue::getSlotValue (this, sceneName ());
}

void TopConfiguration::setSceneLoaderSlot (RecursiveSlotValue *rsv) {
  RecursiveSlotValue::setSlotValue (this, sceneName (), rsv);
}

RecursiveSlotValue *TopConfiguration::getTopLevelSlot () {
  return RecursiveSlotValue::getSlotValue (this, toplevelName ());
}

int TopConfiguration::getStepsPerFrame () const {
  return PositiveIntSlotValue::getSlot (this, stepsPerFrameName ());
}

FFloat TopConfiguration::getTimePerStep () const {
  return PositiveFloatSlotValue::getSlot (this, timePerStepName ());
}

SP<Configuration> TopConfiguration::copy () const {
  return NEW (TopConfiguration (*this));
}

bool TopConfiguration::hasTheSameTypeAs(const Configuration *c) const {
  return 0 != dynamic_cast <const TopConfiguration *> (c);
}

RecursiveSlotValue *TopConfiguration::getIntegrator () {
  return RecursiveSlotValue::getSlotValue (this, integratorName ());
}

const RecursiveSlotValue *TopConfiguration::getIntegrator () const {
  return RecursiveSlotValue::getSlotValue (this, integratorName ());
}

