// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#include "MemoryUtil.h"

// We use malloc/free ourselves, so we can't live with its redefinition in
// MemoryUtil.h.
#undef malloc
#undef free

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef unistd_h_included
#include <unistd.h>
#define unistd_h_included
#endif

#ifndef stdio_h_included
#include <stdio.h>
#define stdio_h_included
#endif

#ifndef NDEBUG
#ifdef MCHECK
#ifndef mcheck_h_included
#include <mcheck.h>
#define mcheck_h_included
#endif
#endif
#endif

// For std::cout, std::cerr.
#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

// For malloc, abort.
#ifndef stdlib_h_included
#include <stdlib.h>
#define stdlib_h_included
#endif

// For strcmp.
#ifndef string_h_included
#include <string.h>
#define string_h_included
#endif

#ifndef Refcount_h_included
#include "Refcount.h"
#endif

#ifndef Dynavec_h_included
#include "Dynavec.h"
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

namespace {
  
  /** The goal of this is to allow human users of the program to discover
      memory leaks that can cause crashes.  The complicating factors are:
      1. Some memory is allocated before we get control.
      2. Some memory is allocated by code we don't control, so we won't get
      line numbers.
      3. Some leaks are deliberately built in to the gcc 3.0.1 runtime for
      efficiency.  See the March 30, 2000 email from Nathan Myers to the
      libstdc++ list archived at
      http://gcc.gnu.org/ml/libstdc++/2000-q1/msg00484.html.

      Thus these strategies for solving the problem are ruled out:

      1. We can't simply report all leaks, because that will lead humans to
      stop reading the output from the leak detector.

      2. We can't simply report only leaks that are from code we control, since
      we could be repeatedly calling some system routine that allocates memory
      on our behalf without calling the corresponding system call that frees
      the memory.

      Instead, we do this:

      Before the start of main, memory allocated will be listed at the end if
      it's code we control.  This way we can avoid writing code that
      contributes to the amount of noise that would emerge from someone who
      simply reports all leaks.
      
      At the beginning of the main routine, main calls the routine
      allLeaksAreInteresting in this file, so we know that all memory allocated
      after this point needs to have leaks reported.

      Then, at the end of main, we call listLeaks.
  */

#define HEADERWORD  0xfeeddeed
#define VIRGINWORD  0xbabeface
#define DELETEDWORD 0xdeadbeef
#define FOOTERWORD  0xfadedebb
  
  // AllocMode says how we allocated it.
  // CMODE means with malloc.
  // CPPSTRUCT means with new.
  // CPPARRAY means with new [].
  // HEADER says it's the list header.
  typedef enum {CMODE, CPPSTRUCT, CPPARRAY, HEADER} AllocMode;

  const char *allocModeToString (AllocMode allocMode) {
    switch (allocMode) {
    case CMODE: return "C memory allocation";
    case CPPSTRUCT: return "C++ structure memory allocation";
    case CPPARRAY: return "C++ array memory allocation";
    case HEADER: return "List header (should never occur)";
    default:
      std::cerr << "Bad allocation mode " << (int) allocMode << "."
                << std::endl;
      abort ();
    }
  }

  struct newLeakInfo {
    int line;
    const char *file;
    newLeakInfo *next;
    newLeakInfo *prev;
    size_t givenSize;
    // true if it was a C-style memory allocation (malloc, calloc, realloc,
    // etc.) and false if it was C++-style (new, delete).  This lets us to
    // enforce the restriction that the two shall not be be mixed.
    AllocMode m_allocMode;
    unsigned int header1;
    // Assume 32 bit ints; have another header word to ensure that the total
    // size of the header is a multiple of 64 bits, so that if malloc is 64-bit
    // aligned then we'll still be 64-bit aligned.
    unsigned int header2;
    // Abort if the header doesn't match.
    inline void checkHeader (const char *gripe) const {
      if (HEADERWORD != header1 || HEADERWORD != header2) {
        std::cerr << gripe << " at " << this << std::endl;
        abort ();
      }
    }
    inline void initHeader () {
      // This would be in a constructor, except I get confused when I try to
      // think about to separate the constructor from memory management enough
      // to avoid tail-eating.  Keep it simple.
      header1 = HEADERWORD;
      header2 = HEADERWORD;
    }
  };
  
  // Compare two of them, for sorting, someday.  For now I'm using
  // sort-lines in an emacs buffer so this code isn't called yet. 
  bool operator> (const newLeakInfo &p1, const newLeakInfo &p2) {
    int cmp = strcmp (p1.file, p2.file);
    if (0 == cmp) {
      if (p1.line > p2.line) {
	return true;
      } else if (p1.line < p2.line) {
	return false;
      } else {
	if (&p1 > &p2) {
	  return true;
	} else {
	  return false;
	}
      }
    } else if (0 < cmp) {
      return false;
    } else {
      assert (0 > cmp);
      return true;
    }
  }

  // Statics are initialized to zero, but don't have a run-time
  // initializer here because this code has to work before all of the
  // run-time initializers have been done, and there's no way for us to
  // guarantee that our static initializer is done before a static initializer
  // that calls new.
  newLeakInfo *listHeader;
  
  size_t roundUpSize (size_t size) {
    // 4 bytes for the footer, 3 bytes for rounding up.
    return ((size + 7) >> 2) << 2;
  }
} // namespace

static void simpleWrite (int fd, const char *string) {
  int status = write (fd, string, strlen (string));
  (void) status; // Suppress unused variable warning for production code.
  // Note that if the assert fails we're going to crash before printing
  // anything, since my assert writes to std::cerr.  However it ought to be
  // possible to use the debugger to figure out what happened.
  assert (status > 0);
}

static void *doNew (size_t size, int line, const char *file,
                    AllocMode allocMode) {
#ifndef NDEBUG
#ifdef MCHECK
  static bool did_mcheck;
  if (!did_mcheck) {
    did_mcheck = 1;
    mcheck (0);
    // If we print here and we are using GCC 3.0.1, we crash.  Apparently std::cerr
    // isn't initialized yet.
    // std::cerr << "Initialized mcheck." << std::endl;
    simpleWrite (2, "Initialized mcheck.\n");
  }
#endif
#endif
  if (MemoryUtil::frequentCheck) {
    MemoryUtil::check ();
  }
  // The number of bytes to actually allocate
  size_t allocSize = roundUpSize (size);
  // Number of ints.
  size_t intSize = allocSize >> 2;
  assert (allocSize >= size);
  assert (0 == allocSize % 4);
  if (! listHeader) {
    listHeader = (newLeakInfo *) malloc (sizeof (newLeakInfo));
    assert (listHeader);
    listHeader->line = -1;
    listHeader->file = 0;
    listHeader->next = listHeader;
    listHeader->prev = listHeader;
    listHeader->givenSize = 0;
    // We won't be looking at the cMode field of the listHeader, but let's
    // avoid uninitialized memory in general.
    listHeader->m_allocMode = HEADER; 
    listHeader->initHeader ();
  }
  newLeakInfo *ptr = (newLeakInfo *) malloc (allocSize+sizeof(newLeakInfo));
  assert (ptr);
  ptr->line = line;
  ptr->file = file;
  // Insert ourselves after listHeader->prev and  before listHeader.
  listHeader->prev->next = ptr;
  ptr->prev = listHeader->prev;
  ptr->next = listHeader;
  ptr->givenSize = size;
  ptr->m_allocMode = allocMode;
  ptr->initHeader ();
  listHeader->prev = ptr;
  int *result = (int *) (ptr + 1);
  result [intSize - 1] = FOOTERWORD;
  for (int i = intSize - 2; i >= 0; i--) {
    result [i] = VIRGINWORD;
  }
  return (void *) result;
}

static bool allInteresting = false;

void MemoryUtil::allLeaksAreInteresting () {
  // Codependent is used here as a description of a psychological problem, not
  // an obscure technial term.
#ifdef CODEPENDENT_MEMORY_LEAK_CHECK
  // We dance around here a little bit so the libstdc++'s preallocated
  // variables will get allocated without being deemed interesting.
  std::cout << "Turning on leak detection.  The answer is " << 42
       << ".  Type something and press return: ";
  cin.clear ();
  int c;
  cin >> c;
  allInteresting = true;
#else
  // Leave allInteresting false.  This way, we won't leak check memory that was
  // not allocated by Fungimol, and if someone links against a different c++
  // library there's some hope of them not getting false alarms.
#endif
}

static void maybeUninteresting (void * const ptr) {
  // delete (0) is permitted by the spec, and the g++ I/O library actually does
  // it.
  if (!ptr) return;
  // If all leaks are interesting, then don't mark this as uninteresting.
  if (allInteresting) return;
  // Splice this element out of the linked list and make it it's own list
  // header.  This will cause delete to function correctly, but it won't be on
  // the list when listLeaks is called.
  newLeakInfo * const toLoop = ((newLeakInfo *) ptr) - 1;
  assert (toLoop->prev);
  assert (toLoop->next);
  toLoop->prev->next = toLoop->next;
  toLoop->next->prev = toLoop->prev;
  toLoop->next = toLoop;
  toLoop->prev = toLoop;
}

// Code I do not control calls the standard new and new[].  Redirect
// them here to have a useless line number and file name, so we can
// still detect leaks and memory corruption and we don't get confused
// later in delete.
void *operator new (size_t size) {
  void * result = doNew (size, __LINE__, __FILE__, CPPSTRUCT);
  maybeUninteresting (result);
  return result;
}

void *operator new [] (size_t size) {
  void *result = doNew (size, __LINE__, __FILE__, CPPARRAY);
  maybeUninteresting (result);
  return result;
}

void *operator new (size_t size, int line, const char *file) {
  return doNew (size, line, file, CPPSTRUCT);
}

void *operator new [] (size_t size, int line, const char *file) {
  return doNew (size, line, file, CPPARRAY);
}

// Whether we've listed the free memory at the "end" by calling listLeaks.  If
// we do deletes after we've listed the free memory, we need to print more
// information to inform the user that the previous list of leaks had too many.
static bool listed = false;

static void doDelete (void * const ptr, const AllocMode expectedMode) {
  // delete (0) is permitted by the spec, and the g++ I/O library actually does
  // it.
  if (!ptr) return;
  if (MemoryUtil::frequentCheck) {
    MemoryUtil::check ();
  }
  const newLeakInfo * const toFree = ((newLeakInfo *) ptr) - 1;
  toFree->checkHeader ("Delete or free of pointer not from new or malloc");
  if (expectedMode != toFree->m_allocMode) {
    const char * expected = allocModeToString (expectedMode);
    const char * actual = allocModeToString (toFree->m_allocMode);
    std::cerr << "A block that was allocated at line " << toFree->line
              << " of " << toFree->file << " with " << actual
              << " was freed with " << expected << ".";
    abort ();
  }
  assert (toFree->prev);
  assert (toFree->next);
  newLeakInfo *oldPrev = toFree->prev;
  newLeakInfo *oldNext = toFree->next;
  
  oldPrev->checkHeader ("Corrupted memory list");
  oldNext->checkHeader ("Corrupted memory list");
  oldNext->prev = oldPrev;
  oldPrev->next = oldNext;
  int freeIntSize =
    (sizeof (newLeakInfo) + roundUpSize (toFree->givenSize)) >> 2;
  unsigned int *toFreeInt = (unsigned int *) toFree;
  if (FOOTERWORD != toFreeInt [freeIntSize - 1]) {
    std::cerr << "Corrupted footer in memory allocated at line "
              << toFree->line << " of "
              << toFree->file << std::endl;
    abort ();
  }
  if (listed) {
    if (listHeader->next == listHeader) {
      printf ("More was freed, and now there are no leaks.");
      listed = false;
    }
    if (toFree->next != toFree) {
      printf ("Freed 0x%lx after listing the leaks.\n", (long) ptr);
    }
  }
  for (int i = 0; i < freeIntSize; i++) {
    toFreeInt [i] = DELETEDWORD;
  }
#if 0
  {
    // The point here is that if you leak everything, then you can
    // put a conditional breakpoint here to stop when a particular
    // address is allocated, and you can be sure that it won't be
    // allocated again.
    static bool complained = false;
    if (!complained) {
      std::cout << "Remember, we've decided to deliberately leak "
        "everything, for debugging. " << std::endl;
      complained = true;
    }
  }
#else
  free ((void *) toFree);
#endif
}

void operator delete (void *ptr) {
  doDelete (ptr, CPPSTRUCT);
}

void operator delete [] (void *x) throw () {
  doDelete (x, CPPARRAY);
}

namespace {
  struct MemoryUtilData
    : public Refcount
  {
    Dynavec <void (*) ()> m_deallocators;
  };
  // With the straightforward code 
  // Data* allDeallocators = 0;
  // we lose if some other class has their static initializers before
  // ours, they register a deallocator, then we get initialized and
  // drop their deallocator.  The more involved code below prevents
  // that.  Same trick was used in FactoryTable.
  MemoryUtilData* initData ();
  MemoryUtilData* allDeallocators = initData ();
  MemoryUtilData* initData () {
    if (allDeallocators) {
      return allDeallocators;
    } else {
      allDeallocators = NEW (MemoryUtilData ());
      allDeallocators->ref();
      return allDeallocators;
    }
  }
  void runDeallocators () {
    if (allDeallocators) {
      for (int i = 0; i < allDeallocators->m_deallocators.size(); i++) {
	allDeallocators->m_deallocators[i] ();
      }
      allDeallocators->deref ();
      assert (0 == allDeallocators->count());
      delete allDeallocators;
      allDeallocators = 0;
    }
  }
}

void MemoryUtil::registerDeallocator (void (*f) ()) {
  initData ();
  allDeallocators->m_deallocators.push (f);
}

void MemoryUtil::listLeaks () {
  runDeallocators ();
  if (listHeader->next == listHeader) {
    std::cout << "No leaks!" << std::endl;
  } else {
    listed = true;
    // Might like to change the if 0 to ifndef NDEBUG.
#ifndef NDEBUG
    std::cout << "Leaks were found!" << std::endl;
#else
    std::cout << "These leaks were found:" << std::endl;
#endif
  }
  // Sort it here, someday.  But not now.
#ifndef NDEBUG
  { // Print.
    newLeakInfo *ptr = listHeader->next;
    for (;;) {
      if (listHeader == ptr) break;
      // The format in the next line lets me do gnu-emacs next-error to
      // find the next leak.
      std::cout << ptr->file << ":" << ptr->line << ": " << ptr->givenSize
                << " bytes allocated at " << (void *) (ptr+1) << "."
                << std::endl;
      ptr = ptr->next;
    }
  }
#endif
}

bool MemoryUtil::frequentCheck = false;

void MemoryUtil::check () {
  newLeakInfo *here = listHeader->next;
  for (;;) {
    if (here == listHeader) break;
    here->checkHeader ("Corrupted header");
    int hereIntSize =
      (sizeof (newLeakInfo) + roundUpSize (here->givenSize)) >> 2;
    unsigned int *hereInt = (unsigned int *) here;
    if (FOOTERWORD != hereInt [hereIntSize - 1]) {
      std::cerr << "Corrupted footer word at " << &hereInt [hereIntSize - 1]
                << std::endl;
      abort ();
    }
    here = here -> next;
  }
}

void *MemoryUtil::fmalloc (int size, int line, const char *file) {
  return doNew (size, line, file, CMODE);  
}

void *MemoryUtil::frealloc (void *oldPtr, int newSize,
                            int line, const char *const file) {
  if (MemoryUtil::frequentCheck) {
    MemoryUtil::check ();
  }
  const newLeakInfo * const info = ((newLeakInfo *) oldPtr) - 1;
  // If we don't check the header on the old block, we may make such a mess
  // when we attempt to copy the number of bytes stored in that header that
  // debugging may become difficult.
  info->checkHeader ("Realloc of bad pointer");
  const int oldSize = info->givenSize;
  char *result = (char *) doNew (newSize, line, file, CMODE);
  for (int i = 0; i < oldSize; i++) {
    result [i] = ((const char *) oldPtr) [i];
  }
  doDelete (oldPtr, CMODE);
  return (void *) result;
}

void MemoryUtil::ffree (void *ptr) {
  doDelete (ptr, CMODE);
}
